@@ -18,7 +18,12 @@ from __future__ import print_function
 import fontforge
 import psMat
 import os
+import re
 
+from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
+from fontTools.misc.py23 import StringIO, tounicode
+from fontTools.ttLib import TTFont
+from pcpp.preprocessor import Preprocessor
 from tempfile import NamedTemporaryFile
 
 def cleanAnchors(font):
@@ -110,14 +115,21 @@ def updateInfo(font, version):
     version = "%07.3f" % version
     font.version = font.version % version
     font.copyright = font.copyright % datetime.now().year
+    font.appendSFNTName("English (US)", "UniqueID", "%s;%s;%s" % (
+        version, font.os2_vendor, font.fontname))
+
+BAD_LOOKUP_FLAG = re.compile(r"(RightToLeft|IgnoreBaseGlyphs|IgnoreLigatures|IgnoreMarks),")
 
 def generateFeatureString(font, lookup):
     with NamedTemporaryFile() as tmp:
         font.generateFeatureFile(tmp.name, lookup)
-        fea = tmp.read()
-        return fea.decode("utf-8")
+        fea = tmp.read().decode("utf-8")
+        # Older versions of FontForge incorrectly seperated lookup flags with
+        # coma.
+        fea = BAD_LOOKUP_FLAG.sub(r"\1", fea)
+        return fea
 
-def generateFeatures(font, feafile):
+def generateFeatures(font, args):
     """Generates feature text by merging feature file with mark positioning
     lookups (already in the font) and making sure they come after kerning
     lookups (from the feature file), which is required by Uniscribe to get
@@ -132,8 +144,16 @@ def generateFeatures(font, feafile):
 
     # open feature file and insert the generated GPOS features in place of the
     # placeholder text
-    with open(feafile, 'rb') as fea:
-        fea_text = fea.read().decode("utf-8")
+    with open(args.features) as f:
+        o = StringIO()
+        preprocessor = Preprocessor()
+        if args.quran:
+            preprocessor.define("QURAN")
+        elif args.slant:
+            preprocessor.define("ITALIC")
+        preprocessor.parse(f)
+        preprocessor.write(o)
+        fea_text = tounicode(o.getvalue(), "utf-8")
     fea_text = fea_text.replace("{%anchors%}", oldfea)
 
     bases = [g.glyphname for g in font.glyphs() if g.glyphclass != "mark"]
@@ -152,16 +172,13 @@ def generateFeatures(font, feafile):
 
     return fea_text
 
-def generateFont(font, feafile, feastring, outfile):
-    from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
-    from fontTools.ttLib import TTFont
-
-    fea = generateFeatures(font, feafile)
+def generateFont(options, font, feastring):
+    fea = generateFeatures(font, args)
     fea += feastring
 
-    flags = ["opentype", "dummy-dsig", "round", "omit-instructions"]
-    if fontforge.version() > '20170805':
-        flags.append("no-mac-names")
+    flags = []
+    if args.output.endswith(".ttf"):
+        flags += ["opentype", "dummy-dsig", "omit-instructions"]
 
     font.selection.all()
     font.correctReferences()
@@ -170,17 +187,35 @@ def generateFont(font, feafile, feastring, outfile):
     # fix some common font issues
     validateGlyphs(font)
 
-    font.generate(outfile, flags=flags)
+    updateInfo(font, args.version)
+
+    font.generate(args.output, flags=flags)
 
     try:
-        ttfont = TTFont(outfile)
+        ttfont = TTFont(args.output)
         addOpenTypeFeaturesFromString(ttfont, fea)
-        ttfont.save(outfile)
+
+        # Filter-out useless Macintosh names
+        name = ttfont["name"]
+        name.names = [n for n in name.names if n.platformID != 1]
+
+        # https://github.com/fontforge/fontforge/pull/3235
+        head = ttfont["head"]
+        # fontDirectionHint is deprecated and must be set to 2
+        head.fontDirectionHint = 2
+        # unset bits 6..10
+        head.flags &= ~0x7e0
+
+        # Drop useless table with timestamp
+        if "FFTM" in ttfont:
+            del ttfont["FFTM"]
+
+        ttfont.save(args.output)
     except:
         with NamedTemporaryFile(delete=False) as tmp:
             tmp.write(fea.encode("utf-8"))
             print("Failed! Inspect temporary file: %r" % tmp.name)
-            os.remove(outfile)
+            os.remove(args.output)
         raise
 
 
@@ -284,11 +319,10 @@ def subsetFont(font, glyphnames, similar=False):
         if glyph.glyphname not in glyphnames:
             font.removeGlyph(glyph)
 
-def subsetFontFT(path, unicodes):
-    from fontTools.ttLib import TTFont
+def subsetFontFT(path, unicodes, quran=False):
     from fontTools import subset
 
-    font = TTFont(path)
+    font = TTFont(path, recalcTimestamp=False)
 
     options = subset.Options()
     options.set(layout_features='*', name_IDs='*', name_languages='*',
@@ -297,6 +331,9 @@ def subsetFontFT(path, unicodes):
     subsetter.populate(unicodes=unicodes)
     subsetter.subset(font)
 
+    if quran:
+        font["OS/2"].sTypoAscender = font["hhea"].ascent = font["head"].yMax
+
     font.save(path)
 
 def makeNumerators(font):
@@ -445,6 +482,12 @@ def mergeLatin(font, italic=False, glyphs=None, quran=False):
             latinfont.selection.select(name)
             latinfont.paste()
 
+            # Fixup for old FontForge that also renamed any other glyph that
+            # started with name!
+            for other in latinfont.glyphs():
+                if other.glyphname.startswith(glyph.glyphname) and other != glyph:
+                    other.glyphname = other.glyphname.replace(glyph.glyphname, name)
+
             rtl = latinfont.createChar(-1, name + ".rtl")
             rtl.addReference(name, italic)
             rtl.useRefsMetrics(name)
@@ -510,13 +553,12 @@ def mergeLatin(font, italic=False, glyphs=None, quran=False):
 
     return fea
 
-def makeSlanted(infile, outfile, feafile, version, slant):
-
-    font = makeDesktop(infile, outfile, feafile, version, False)
+def makeSlanted(options):
+    font = makeDesktop(options, False)
 
     # compute amout of skew, magic formula copied from fontforge sources
     import math
-    skew = psMat.skew(-slant * math.pi/180.0)
+    skew = psMat.skew(-options.slant * math.pi/180.0)
 
     # Remove Arabic math alphanumerics, they are upright-only.
     font.selection.select(["ranges"], "u1EE00", "u1EEFF")
@@ -524,7 +566,7 @@ def makeSlanted(infile, outfile, feafile, version, slant):
         font.removeGlyph(glyph)
 
     font.selection.all()
-    punct = ("exclam", "period", "guillemotleft", "guillemotright",
+    punct = ("exclam", "period.ara", "guillemotleft.ara", "guillemotright.ara",
              "braceleft", "bar", "braceright", "bracketleft", "bracketright",
              "parenleft", "parenright", "slash", "backslash", "brokenbar",
              "uni061F", "dot.1", "dot.2")
@@ -535,7 +577,7 @@ def makeSlanted(infile, outfile, feafile, version, slant):
     font.transform(skew)
 
     # fix metadata
-    font.italicangle = slant
+    font.italicangle = options.slant
     font.fullname += " Slanted"
     if font.weight == "Bold":
         font.fontname = font.fontname.replace("Bold", "BoldSlanted")
@@ -545,7 +587,7 @@ def makeSlanted(infile, outfile, feafile, version, slant):
 
     fea = mergeLatin(font, italic=skew)
     makeNumerators(font)
-    generateFont(font, feafile, fea, outfile)
+    generateFont(options, font, fea)
 
 def scaleGlyph(glyph, amount):
     """Scales the glyph, but keeps it centered around its original bounding
@@ -568,8 +610,8 @@ def scaleGlyph(glyph, amount):
     if width == 0:
         glyph.width = width
 
-def makeQuran(infile, outfile, feafile, version):
-    font = makeDesktop(infile, outfile, feafile, version, False)
+def makeQuran(options):
+    font = makeDesktop(options, False)
 
     # fix metadata
     font.fontname = font.fontname.replace("-Regular", "Quran-Regular")
@@ -584,23 +626,10 @@ def makeQuran(infile, outfile, feafile, version):
 
     fea = mergeLatin(font, glyphs=digits, quran=True)
 
-    punct = ("period", "guillemotleft", "guillemotright", "braceleft", "bar",
-             "braceright", "bracketleft", "bracketright", "parenleft",
-             "parenright", "slash", "backslash")
-
-    for name in punct:
-        if name+".ara" in font:
-            glyph = font[name+".ara"]
-            glyph.glyphname = name
-            glyph.unicode = fontforge.unicodeFromName(name)
-
-    # abuse U+065C as a below form of U+06EC, for Qaloon
-    dotabove = font["uni06EC"]
-    dotbelow = font["uni065C"]
-    delta = dotbelow.boundingBox()[-1] - dotabove.boundingBox()[-1]
-    dotbelow.references = []
-    dotbelow.addReference(dotabove.glyphname, psMat.translate(0, delta))
-    dotbelow.addAnchorPoint("TashkilTashkilBelow", "basemark", 220, dotbelow.boundingBox()[1] - 100)
+    for glyph in font.glyphs():
+        if glyph.glyphname.endswith(".ara"):
+            glyph.glyphname = glyph.glyphname[:-4]
+            glyph.unicode = fontforge.unicodeFromName(glyph.glyphname)
 
     # scale some vowel marks and dots down a bit
     scaleGlyph(font["uni0651"], 0.8)
@@ -613,62 +642,41 @@ def makeQuran(infile, outfile, feafile, version):
     # vertically at the level of the base of waqf marks
     fea += makeQuranSajdaLine(font, font[0x06D7].boundingBox()[1])
 
-    quran_glyphs = []
-    quran_glyphs += digits
-    quran_glyphs += punct
-    quran_glyphs += ("space",
-            "uni060C", "uni0615", "uni0617", "uni0618", "uni0619", "uni061A",
-            "uni061B", "uni061E", "uni061F", "uni0621", "uni0622", "uni0623",
-            "uni0624", "uni0625", "uni0626", "uni0627", "uni0628", "uni0629",
-            "uni062A", "uni062B", "uni062C", "uni062D", "uni062E", "uni062F",
-            "uni0630", "uni0631", "uni0632", "uni0633", "uni0634", "uni0635",
-            "uni0636", "uni0637", "uni0638", "uni0639", "uni063A", "uni0640",
-            "uni0641", "uni0642", "uni0643", "uni0644", "uni0645", "uni0646",
-            "uni0647", "uni0648", "uni0649", "uni064A", "uni064B", "uni064C",
-            "uni064D", "uni064E", "uni064F", "uni0650", "uni0651", "uni0652",
-            "uni0653", "uni0654", "uni0655", "uni0656", "uni0657", "uni0658",
-            "uni065C", "uni0660", "uni0661", "uni0662", "uni0663", "uni0664",
-            "uni0665", "uni0666", "uni0667", "uni0668", "uni0669", "uni066E",
-            "uni066F", "uni06A1", "uni06BA", "uni0670", "uni0671", "uni067A",
-            "uni06CC", "uni06D6", "uni06D7", "uni06D8", "uni06D9", "uni06DA",
-            "uni06DB", "uni06DC", "uni06DD", "uni06DE", "uni06DF", "uni06E0",
-            "uni06E1", "uni06E2", "uni06E3", "uni06E4", "uni06E5", "uni06E6",
-            "uni06E7", "uni06E8", "uni06E9", "uni06EA", "uni06EB", "uni06EC",
-            "uni06ED", "uni06F0", "uni06F1", "uni06F2", "uni06F3", "uni06F4",
-            "uni06F5", "uni06F6", "uni06F7", "uni06F8", "uni06F9", "uni08F0",
-            "uni08F1", "uni08F2", "uni2000", "uni2001", "uni2002", "uni2003",
-            "uni2004", "uni2005", "uni2006", "uni2007", "uni2008", "uni2009",
-            "uni200A", "uni200B", "uni200C", "uni200D", "uni200E", "uni200F",
-            "uni2028", "uni2029", "uni202A", "uni202B", "uni202C", "uni202D",
-            "uni202E", "uni202F", "uni25CC", "uniFD3E", "uniFD3F", "uniFDFA",
-            "uniFDFD")
-    quran_glyphs += ("uni030A", "uni0325") # ring above and below
-    quran_glyphs += ["uni0305"] # overline
-    unicodes = [font[n].unicode for n in quran_glyphs]
+    unicodes = [font[n].unicode for n in digits]
+    unicodes += ['.', '(', ')', '[', ']', '{', '}', '|', ' ', '/', '\\',
+                 0x00A0,
+                 0x00AB, 0x00BB, 0x0305, 0x030A, 0x0325, 0x060C, 0x0615,
+                 0x0617, 0x0618, 0x0619, 0x061A, 0x061B, 0x061E, 0x061F,
+                 0x0621, 0x0622, 0x0623, 0x0624, 0x0625, 0x0626, 0x0627,
+                 0x0628, 0x0629, 0x062A, 0x062B, 0x062C, 0x062D, 0x062E,
+                 0x062F, 0x0630, 0x0631, 0x0632, 0x0633, 0x0634, 0x0635,
+                 0x0636, 0x0637, 0x0638, 0x0639, 0x063A, 0x0640, 0x0641,
+                 0x0642, 0x0643, 0x0644, 0x0645, 0x0646, 0x0647, 0x0648,
+                 0x0649, 0x064A, 0x064B, 0x064C, 0x064D, 0x064E, 0x064F,
+                 0x0650, 0x0651, 0x0652, 0x0653, 0x0654, 0x0655, 0x0656,
+                 0x0657, 0x0658, 0x065C, 0x0660, 0x0661, 0x0662, 0x0663,
+                 0x0664, 0x0665, 0x0666, 0x0667, 0x0668, 0x0669, 0x066E,
+                 0x066F, 0x0670, 0x0671, 0x067A, 0x06A1, 0x06BA, 0x06CC,
+                 0x06D6, 0x06D7, 0x06D8, 0x06D9, 0x06DA, 0x06DB, 0x06DC,
+                 0x06DD, 0x06DE, 0x06DF, 0x06E0, 0x06E1, 0x06E2, 0x06E3,
+                 0x06E4, 0x06E5, 0x06E6, 0x06E7, 0x06E8, 0x06E9, 0x06EA,
+                 0x06EB, 0x06EC, 0x06ED, 0x06F0, 0x06F1, 0x06F2, 0x06F3,
+                 0x06F4, 0x06F5, 0x06F6, 0x06F7, 0x06F8, 0x06F9, 0x08F0,
+                 0x08F1, 0x08F2, 0x08F3, 0x2000, 0x2001, 0x2002, 0x2003,
+                 0x2004,
+                 0x2005, 0x2006, 0x2007, 0x2008, 0x2009, 0x200A, 0x200B,
+                 0x200C, 0x200D, 0x200E, 0x200F, 0x2028, 0x2029, 0x202A,
+                 0x202B, 0x202C, 0x202D, 0x202E, 0x202F, 0x25CC, 0xFD3E,
+                 0xFD3F, 0xFDFA, 0xFDFD]
+    unicodes = [isinstance(u, str) and ord(u) or u for u in unicodes]
 
-    # set font ascent to the highest glyph in the font so that waqf marks don't
-    # get truncated
-    # we could have set os2_typoascent_add and hhea_ascent_add, but ff makes
-    # the offset relative to em-size in the former and font bounds in the
-    # later, but we want both to be relative to font bounds
-    ymax = 0
-    for name in quran_glyphs:
-        glyph = font[name]
-        bb = glyph.boundingBox()
-        if bb[-1] > ymax:
-            ymax = bb[-1]
+    generateFont(options, font, fea)
+    subsetFontFT(options.output, unicodes, quran=True)
 
-    font.os2_typoascent = font.hhea_ascent = ymax
-
-    generateFont(font, feafile, fea, outfile)
-    subsetFontFT(outfile, unicodes)
-
-def makeDesktop(infile, outfile, feafile, version, generate=True):
-    font = fontforge.open(infile)
+def makeDesktop(options, generate=True):
+    font = fontforge.open(options.input)
     font.encoding = "UnicodeFull" # avoid a crash if compact was set
 
-    updateInfo(font, version)
-
     # remove anchors that are not needed in the production font
     cleanAnchors(font)
 
@@ -680,7 +688,7 @@ def makeDesktop(infile, outfile, feafile, version, generate=True):
     if generate:
         fea = mergeLatin(font)
         makeNumerators(font)
-        generateFont(font, feafile, fea, outfile)
+        generateFont(options, font, fea)
     else:
         return font
 
@@ -697,8 +705,8 @@ if __name__ == "__main__":
     args = parser.parse_args()
 
     if args.slant:
-        makeSlanted(args.input, args.output, args.features, args.version, args.slant)
+        makeSlanted(args)
     elif args.quran:
-        makeQuran(args.input, args.output, args.features, args.version)
+        makeQuran(args)
     else:
-        makeDesktop(args.input, args.output, args.features, args.version)
+        makeDesktop(args)
