@@ -50,6 +50,7 @@ MARKS_GLYPHS = (
     "uni08F0",
     "uni08F1",
     "uni08F2",
+    "uni08F3",
     "hamza.wasl",
     "Dot",
     "TwoDots",
@@ -92,6 +93,8 @@ GROUPS = {
     PAUSES_GLYPHS: BLUE,
 }
 
+HASHES = {}
+
 def newLayer(name, colorID):
     return getTableModule("COLR").LayerRecord(name=name, colorID=colorID)
 
@@ -106,9 +109,14 @@ def getGlyphColor(glyphName):
 def colorize(font):
     COLR = newTable("COLR")
     CPAL = newTable("CPAL")
-    glyf = font["glyf"]
     hmtx = font["hmtx"]
 
+    glyf = font.get("glyf", None)
+    CFF  = font.get("CFF ", None)
+
+    if glyf is None:
+        assert False, "CFF is not supported"
+
     CPAL.version = 0
     COLR.version = 0
 
@@ -141,8 +149,16 @@ def colorize(font):
                     # identical layers and avoid needless duplication.
                     width = hmtx[name][0]
                     lsb = hmtx[componentName][1] + trans[4]
-                    identifier = hash((trans, width, lsb))
-                    newName = "%s.%s" % (componentName, identifier)
+                    componentHash = hash((trans, width, lsb))
+
+                    if componentName not in HASHES:
+                        HASHES[componentName] = []
+
+                    if componentHash not in HASHES[componentName]:
+                        HASHES[componentName].append(componentHash)
+
+                    index = HASHES[componentName].index(componentHash)
+                    newName = "%s.l%s" % (componentName, index)
 
                     if newName not in font.glyphOrder:
                         font.glyphOrder.append(newName)
@@ -169,16 +185,12 @@ def colorize(font):
 
 def rename(font):
     for name in font["name"].names:
-        if name.nameID in (1, 4, 6):
+        if name.nameID in (1, 3, 4, 6):
             string = name.toUnicode()
-            if name.nameID == 6:
-                if "-" in string:
-                    family, subfamily = string.split("-")
-                    string = "%sColored-%s" % (family, subfamily)
-                else:
-                    string += "Colored"
+            if name.nameID in (3, 6):
+                string = string.replace("Quran", "QuranColored")
             else:
-                string += " Colored"
+                string = string.replace("Quran", "Quran Colored")
 
             name.string = string.encode(name.getEncoding())
 
@@ -189,7 +201,7 @@ def main():
 
     args = parser.parse_args()
 
-    font = TTFont(args.infile)
+    font = TTFont(args.infile, recalcTimestamp=False)
 
     colorize(font)
     rename(font)
